import numpy as np
import torch

import thelper


def test_accuracy_val():
    batch_size = 32
    num_classes = 10
    target_count = 128
    np.random.seed()
    targets = np.asarray([idx % num_classes for idx in range(target_count)])
    preds = np.random.rand(128, 10)
    nb_correct = np.count_nonzero(targets == np.argmax(preds, axis=1))
    metric = thelper.optim.metrics.Accuracy()
    iter_count = target_count // batch_size
    for iter_idx in range(iter_count):
        metric.update(task=None, input=None,
                      pred=torch.from_numpy(preds[iter_idx * batch_size:(iter_idx + 1) * batch_size]),
                      target=torch.from_numpy(targets[iter_idx * batch_size:(iter_idx + 1) * batch_size]),
                      sample=None, loss=None, iter_idx=iter_idx, max_iters=iter_count, epoch_idx=0, max_epochs=1)
    assert np.isclose(metric.eval(), (nb_correct / target_count) * 100)
    # try looping over, max window should be just the right size
    for iter_idx in range(iter_count):
        metric.update(task=None, input=None,
                      pred=torch.from_numpy(preds[iter_idx * batch_size:(iter_idx + 1) * batch_size]),
                      target=torch.from_numpy(targets[iter_idx * batch_size:(iter_idx + 1) * batch_size]),
                      sample=None, loss=None, iter_idx=iter_idx, max_iters=iter_count, epoch_idx=0, max_epochs=1)
    assert np.isclose(metric.eval(), (nb_correct / target_count) * 100)


def test_accuracy_1d(mocker):
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    class_count = 10
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    metric_config = {"metric": {
        "type": "thelper.optim.metrics.Accuracy",
        "params": {"top_k": 3}
    }}
    metrics = thelper.optim.create_metrics(metric_config)
    metric = metrics["metric"]
    assert isinstance(metric, thelper.optim.metrics.Accuracy)
    assert metric.goal == thelper.optim.Metric.maximize
    assert metric.top_k == 3
    with mocker.patch.object(thelper.optim.metrics.logger, "warning"):
        assert metric.eval() == 0.0
    assert repr(metric)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size,)))
        preds.append(torch.rand((curr_batch_size, class_count)))
        metric.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                      {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                      None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    res = metric.eval()
    assert res is not None and isinstance(res, float) and 0 <= res <= 100
    metric.reset()
    assert metric.eval() == 0.0
    tot_idx = 0
    for iter_idx in range(iter_count):
        metric.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                      {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                      None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert metric.eval() == res


def test_accuracy_nd(mocker):
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    output_shape = (16, 16)
    class_count = 10
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    metric_config = {"metric": {
        "type": "thelper.optim.metrics.Accuracy",
        "params": {"top_k": 3}
    }}
    metrics = thelper.optim.create_metrics(metric_config)
    metric = metrics["metric"]
    assert isinstance(metric, thelper.optim.metrics.Accuracy)
    assert metric.goal == thelper.optim.Metric.maximize
    assert metric.top_k == 3
    with mocker.patch.object(thelper.optim.metrics.logger, "warning"):
        assert metric.eval() == 0.0
    assert repr(metric)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size, *output_shape)))
        preds.append(torch.rand((curr_batch_size, class_count, *output_shape)))
        metric.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                      {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                      None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    res = metric.eval()
    assert res is not None and isinstance(res, float) and 0 <= res <= 100
    metric.reset()
    assert metric.eval() == 0.0
    tot_idx = 0
    for iter_idx in range(iter_count):
        metric.update(task, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                      {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                      None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert metric.eval() == res


def test_mae_mse(mocker):
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    metric_config = {"mae": {"type": "thelper.optim.metrics.MeanAbsoluteError"},
                     "mse": {"type": "thelper.optim.metrics.MeanSquaredError"}}
    metrics = thelper.optim.create_metrics(metric_config)
    mae, mse = metrics["mae"], metrics["mse"]
    assert isinstance(mae, thelper.optim.metrics.MeanAbsoluteError)
    assert isinstance(mse, thelper.optim.metrics.MeanSquaredError)
    assert mae.goal == thelper.optim.Metric.minimize
    assert mse.goal == thelper.optim.Metric.minimize
    with mocker.patch.object(thelper.optim.metrics.logger, "warning"):
        assert mae.eval() == 0.0
        assert mse.eval() == 0.0
    assert repr(mae) and repr(mse)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.randn((curr_batch_size,)))
        preds.append(torch.randn((curr_batch_size, )))
        mae.update(None, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                   {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                   None, iter_idx, iter_count, 0, 1)
        mse.update(None, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                   {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                   None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    mae_res, mse_res = mae.eval(), mse.eval()
    assert mae_res is not None and isinstance(mae_res, float) and mae_res >= 0
    assert mse_res is not None and isinstance(mse_res, float) and mse_res >= 0
    assert mse_res >= mae_res
    mae.reset()
    mse.reset()
    assert mae.eval() == 0.0
    assert mse.eval() == 0.0
    tot_idx = 0
    for iter_idx in range(iter_count):
        mae.update(None, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                   {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                   None, iter_idx, iter_count, 0, 1)
        mse.update(None, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                   {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                   None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert mae.eval() == mae_res
    assert mse.eval() == mse_res


def test_external_metrics():
    batch_size = 32
    iter_count = 64
    class_count = 7
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    metric_config = {
        "f1": {
            "type": "thelper.optim.metrics.ExternalMetric",
            "params": {
                "metric_name": "sklearn.metrics.f1_score",
                "metric_params": {},
                "metric_type": "classif_top1",
                "target_name": "0",
                "metric_goal": "max"
            }
        },
        "auc": {
            "type": "thelper.optim.metrics.ExternalMetric",
            "params": {
                "metric_name": "sklearn.metrics.roc_auc_score",
                "metric_params": {},
                "metric_type": "classif_scores",
                "target_name": "1",
                "metric_goal": "max",
                "class_names": class_names
            }
        },
    }
    metrics = thelper.optim.create_metrics(metric_config)
    metric_f1, metric_auc = metrics["f1"], metrics["auc"]
    assert isinstance(metric_f1, thelper.optim.metrics.ExternalMetric)
    assert isinstance(metric_auc, thelper.optim.metrics.ExternalMetric)
    assert metric_f1.goal == thelper.optim.Metric.maximize
    assert metric_auc.goal == thelper.optim.Metric.maximize
    assert repr(metric_f1) and repr(metric_auc)
    targets, preds = [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size,)))
        preds.append(torch.rand((curr_batch_size, class_count)))
        metric_f1.update(task, None, preds[iter_idx], targets[iter_idx],
                         {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                         None, iter_idx, iter_count, 0, 1)
        metric_auc.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    f1_res, auc_res = metric_f1.eval(), metric_auc.eval()
    assert f1_res is not None and isinstance(f1_res, float) and 0 <= f1_res <= 1
    assert auc_res is not None and isinstance(auc_res, float) and 0 <= auc_res <= 1
    metric_f1.reset()
    tot_idx = 0
    for iter_idx in range(iter_count):
        metric_f1.update(task, None, preds[iter_idx], targets[iter_idx],
                         {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                         None, iter_idx, iter_count, 0, 1)
        metric_auc.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                          None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert metric_f1.eval() == f1_res
    assert metric_auc.eval() == auc_res


def test_roccurve():
    batch_size = 32
    iter_count = 128
    class_count = 3
    class_names = [str(i) for i in range(class_count)]
    task = thelper.tasks.Classification(class_names, "input", "gt", ["idx"])
    metric_config = {
        "tpr": {
            "type": "thelper.optim.metrics.ROCCurve",
            "params": {
                "target_name": "1",
                "target_fpr": 0.99
            }
        },
        "fpr": {
            "type": "thelper.optim.metrics.ROCCurve",
            "params": {
                "target_name": "0",
                "target_tpr": 0.9
            }
        },
        "auc": {
            "type": "thelper.optim.metrics.ROCCurve",
            "params": {
                "target_name": "2"
            }
        },
    }
    metrics = thelper.optim.create_metrics(metric_config)
    metric_tpr, metric_fpr, metric_auc = metrics["tpr"], metrics["fpr"], metrics["auc"]
    assert isinstance(metric_tpr, thelper.optim.metrics.ROCCurve)
    assert isinstance(metric_fpr, thelper.optim.metrics.ROCCurve)
    assert isinstance(metric_auc, thelper.optim.metrics.ROCCurve)
    assert metric_tpr.goal == thelper.optim.Metric.maximize
    assert metric_fpr.goal == thelper.optim.Metric.minimize
    assert metric_auc.goal == thelper.optim.Metric.maximize
    assert repr(metric_tpr) and repr(metric_fpr) and repr(metric_auc)
    metric_ref = thelper.optim.metrics.ExternalMetric(metric_name="sklearn.metrics.roc_auc_score",
                                                      metric_type="classif_scores",
                                                      target_name="2", metric_goal="max")
    targets, preds = [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        targets.append(torch.randint(low=0, high=class_count, size=(curr_batch_size,)))
        preds.append(torch.rand((curr_batch_size, class_count)))
        metric_tpr.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        metric_fpr.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        metric_auc.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        metric_ref.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    tpr_res, fpr_res, auc_res, ref_res = metric_tpr.eval(), metric_fpr.eval(), metric_auc.eval(), metric_ref.eval()

    assert tpr_res is not None and isinstance(tpr_res, float) and 0 <= tpr_res <= 1
    assert fpr_res is not None and isinstance(fpr_res, float) and 0 <= fpr_res <= 1
    assert auc_res is not None and isinstance(auc_res, float) and 0 <= auc_res <= 1
    assert ref_res is not None and isinstance(ref_res, float) and 0 <= ref_res <= 1
    assert np.isclose(ref_res, auc_res)
    metric_tpr.reset()
    metric_fpr.reset()
    metric_auc.reset()
    tot_idx = 0
    for iter_idx in range(iter_count):
        metric_tpr.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        metric_fpr.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        metric_auc.update(task, None, preds[iter_idx], targets[iter_idx],
                          {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                          None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert metric_tpr.eval() == tpr_res
    assert metric_fpr.eval() == fpr_res
    assert metric_auc.eval() == auc_res


def test_psnr(mocker):
    batch_size = 16
    iter_count = 32
    input_shape = (3, 32, 32)
    metric_config = {"psnr": {"type": "thelper.optim.metrics.PSNR"}}
    metric_psnr = thelper.optim.create_metrics(metric_config)["psnr"]
    assert isinstance(metric_psnr, thelper.optim.metrics.PSNR)
    assert metric_psnr.goal == thelper.optim.Metric.maximize
    with mocker.patch.object(thelper.optim.metrics.logger, "warning"):
        assert metric_psnr.eval() == 0.0
    assert repr(metric_psnr)
    inputs, targets, preds = [], [], []
    tot_idx = 0
    for iter_idx in range(iter_count):
        # set batch size to one for 'lingering' sample in last minibatch
        curr_batch_size = batch_size if iter_idx < iter_count - 1 else 1
        inputs.append(torch.randn((curr_batch_size, *input_shape)))
        targets.append(torch.rand((curr_batch_size,)))
        preds.append(torch.rand((curr_batch_size,)))
        metric_psnr.update(None, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                           {"idx": [tot_idx + idx for idx in range(curr_batch_size)]},
                           None, iter_idx, iter_count, 0, 1)
        tot_idx += curr_batch_size
    psnr_res = metric_psnr.eval()
    assert psnr_res is not None and isinstance(psnr_res, float) and psnr_res >= 0
    metric_psnr.reset()
    assert metric_psnr.eval() == 0.0
    tot_idx = 0
    for iter_idx in range(iter_count):
        metric_psnr.update(None, inputs[iter_idx], preds[iter_idx], targets[iter_idx],
                           {"idx": [tot_idx + idx for idx in range(targets[iter_idx].shape[0])]},
                           None, iter_idx, iter_count, 0, 1)
        tot_idx += targets[iter_idx].shape[0]
    assert metric_psnr.eval() == psnr_res




import thelper.nn


class SRCNN(thelper.nn.Module):
    """Implements the SRCNN architecture.

    See Dong et al., "Image Super-Resolution Using Deep Convolutional Networks" (2014) for more
    information (https://arxiv.org/abs/1501.00092).
    """

    def __init__(self, task, num_channels=1, base_filter=64, groups=1):
        # note: must always forward args to base class to keep backup
        super(SRCNN, self).__init__(task, num_channels=num_channels, base_filter=base_filter, groups=groups)
        self.conv1 = thelper.nn.common.ConvBlock(num_channels, base_filter * groups, kernel_size=9,
                                                 stride=1, padding=0, activation="relu", norm=None, groups=groups)
        self.conv2 = thelper.nn.common.ConvBlock(base_filter * groups, base_filter // 2 * groups, kernel_size=5,
                                                 stride=1, padding=0, activation="relu", norm=None, groups=groups)
        self.conv3 = thelper.nn.common.ConvBlock((base_filter // 2) * groups, num_channels, kernel_size=5,
                                                 stride=1, padding=0, activation=None, norm=None, groups=groups)
        self.set_task(task)

    def forward(self, x):
        x0 = x.view(x.shape[0] * x.shape[1], 1, x.shape[2], x.shape[3])
        x0 = self.conv1(x0)
        x0 = self.conv2(x0)
        x0 = self.conv3(x0)
        x0 = x0.view(x.shape[0], x.shape[1], x0.shape[2], x0.shape[3])
        return x0

    def weight_init(self):
        for m in self.modules():
            thelper.nn.common.weights_init_xavier(m)

    def set_task(self, task):
        if not isinstance(task, thelper.tasks.Regression):
            raise AssertionError("SRCNN architecture only available for super res regression tasks")
        self.task = task



"""Segmentation task interface module.

This module contains a class that defines the objectives of models/trainers for segmentation tasks.
"""
import logging
from typing import Optional  # noqa: F401

import numpy as np
import PIL.Image
import torch
import tqdm

import thelper.concepts
import thelper.utils
from thelper.ifaces import ClassNamesHandler
from thelper.tasks.utils import Task

logger = logging.getLogger(__name__)


@thelper.concepts.segmentation
class Segmentation(Task, ClassNamesHandler):
    """Interface for pixel-level labeling/classification (segmentation) tasks.

    This specialization requests that when given an input tensor, the trained model should
    provide prediction scores for each predefined label (or class), for each element of the
    input tensor. The label names are used here to help categorize samples, and to assure that
    two tasks are only identical when their label counts and ordering match.

    Attributes:
        class_names: map of class name-value pairs to predict for each pixel.
        input_key: the key used to fetch input tensors from a sample dictionary.
        label_map_key: the key used to fetch label (class) maps from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.
        dontcare: value of the 'dontcare' label (if any) used in the class map.
        color_map: map of class name-color pairs to use when displaying results.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.train.segm.ImageSegmTrainer`
    """

    def __init__(self, class_names, input_key, label_map_key,
                 meta_keys=None, dontcare=None, color_map=None):
        """Receives and stores the class (or label) names to predict, the input tensor key,
        the groundtruth label (class) map key, the extra (meta) keys produced by the dataset
        parser(s), the 'dontcare' label value that might be present in gt maps (if any), and
        the color map used to swap label indices for colors when displaying results.

        The class names can be provided as a list of strings, as a path to a json file that
        contains such a list, or as a map of predefined name-value pairs to use in gt maps.
        This list/map must contain at least two elements. All other arguments are used as-is
        to index dictionaries, and must therefore be key-compatible types.
        """
        super(Segmentation, self).__init__(input_key, label_map_key, meta_keys)
        if "dontcare" in class_names:
            logger.warning("found reserved 'dontcare' label in input classes; it will be removed from the internal list")
            if isinstance(class_names, list):
                del class_names[class_names.index("dontcare")]
            elif isinstance(class_names, dict):
                del class_names["dontcare"]
        ClassNamesHandler.__init__(self, class_names=class_names)
        self.dontcare = dontcare
        self.color_map = color_map

    @property
    def dontcare(self):
        """Returns the 'dontcare' label value used in loss functions (can be ``None``)."""
        return self._dontcare

    @dontcare.setter
    def dontcare(self, dontcare):
        """Sets the 'dontcare' label value for this segmentation task (can be ``None``)."""
        if dontcare is not None:
            assert isinstance(dontcare, int), "'dontcare' value should be integer (index)"
            assert dontcare not in self.class_indices.values(), "found 'dontcare' value tied to another class label"
        self._dontcare = dontcare

    @property
    def color_map(self):
        """Returns the color map used to swap label indices for colors when displaying results."""
        return self._color_map

    @color_map.setter
    def color_map(self, color_map):
        """Sets the color map used to swap label indices for colors when displaying results."""
        if color_map is not None:
            assert isinstance(color_map, dict), "color map should be given as dictionary"
            self._color_map = {}
            assert all([isinstance(k, int) for k in color_map]) or all([isinstance(k, str) for k in color_map]), \
                "color map keys should be only class names or only class indices"
            for key, val in color_map.items():
                if isinstance(key, str):
                    if key == "dontcare" and self.dontcare is not None:
                        key = self.dontcare
                    else:
                        assert key in self.class_indices, f"could not find color map key '{key}' in class names"
                        key = self.class_indices[key]
                assert key in self.class_indices.values() or key == self.dontcare, f"unrecognized class index '{key}'"
                if isinstance(val, (list, tuple)):
                    val = np.asarray(val)
                assert isinstance(val, np.ndarray) and val.size == 3, "color values should be given as triplets"
                self._color_map[key] = val
            if self.dontcare is not None and self.dontcare not in self._color_map:
                self._color_map[self.dontcare] = np.asarray([0, 0, 0])  # use black as default 'dontcare' color
        else:
            self._color_map = {}

    def get_class_sizes(self, samples):
        """Given a list of samples, returns a map of element counts for each class label."""
        assert samples is not None and samples, "provided invalid sample list"
        elem_counts = {class_name: 0 for class_name in self.class_names}
        if self.dontcare is not None:
            elem_counts["dontcare"] = 0
        warned_unknown_value_flag = False
        for sample_idx, sample in tqdm.tqdm(enumerate(samples), desc="cumulating label counts", total=len(samples)):
            if self.gt_key is None or self.gt_key not in sample:
                continue
            else:
                labels = sample[self.gt_key]
                if isinstance(labels, torch.Tensor):
                    labels = labels.cpu().numpy()
                if isinstance(labels, PIL.Image.Image):
                    labels = np.array(labels)
                assert isinstance(labels, np.ndarray), "unsupported label map type ('%s')" % str(type(labels))
                # here, we assume labels are given as some integer type that corresponds to class name indices
                curr_elem_counts = {cname: np.count_nonzero(labels == cval) for cname, cval in self.class_indices.items()}
                dontcare_elem_count = 0 if self.dontcare is None else np.count_nonzero(labels == self.dontcare)
                if (sum(curr_elem_counts.values()) + dontcare_elem_count) != labels.size and not warned_unknown_value_flag:
                    logger.warning("some label maps contain values that are unknown (i.e. with no proper class mapping)")
                    warned_unknown_value_flag = True
                for class_name in self.class_names:
                    elem_counts[class_name] += curr_elem_counts[class_name]
                if dontcare_elem_count > 0:
                    elem_counts["dontcare"] += dontcare_elem_count
        return elem_counts

    def check_compat(self, task, exact=False):
        # type: (Segmentation, Optional[bool]) -> bool
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. If ``exact = True``, all fields will be checked for exact (perfect)
        compatibility (in this case, matching meta keys and class maps).
        """
        if isinstance(task, Segmentation):
            # if both tasks are related to segmentation: gt keys, class names, and dc must match
            return self.input_key == task.input_key and self.dontcare == task.dontcare and \
                (self.gt_key is None or task.gt_key is None or self.gt_key == task.gt_key) and \
                all([cls in self.class_names for cls in task.class_names]) and \
                (not exact or (self.class_names == task.class_names and
                               set(self.meta_keys) == set(task.meta_keys) and
                               self.color_map.keys() == task.color_map.keys() and
                               all([np.array_equal(self.color_map[k], task.color_map[k]) for k in self.color_map]) and
                               self.gt_key == task.gt_key))
        elif type(task) == Task:
            # if 'task' simply has no gt, compatibility rests on input key only
            return not exact and self.input_key == task.input_key and task.gt_key is None
        return False

    def get_compat(self, task):
        """Returns a task instance compatible with the current task and the given one."""
        assert isinstance(task, Segmentation) or type(task) == Task, \
            f"cannot create compatible task from types '{type(task)}' and '{type(self)}'"
        if isinstance(task, Segmentation):
            assert self.input_key == task.input_key, "input key mismatch, cannot create compatible task"
            assert self.gt_key is None or task.gt_key is None or self.gt_key == task.gt_key, \
                "gt key mismatch, cannot create compatible task"
            assert self.dontcare == task.dontcare, "dontcare value mismatch, cannot create compatible task"
            meta_keys = list(set(self.meta_keys + task.meta_keys))
            # cannot use set for class names, order needs to stay intact!
            class_indices = {cname: cval for cname, cval in task.class_indices.items() if cname not in self.class_indices}
            class_indices = {**self.class_indices, **class_indices}
            color_map = {cname: cval for cname, cval in task.color_map.items() if cname not in self.color_map}
            color_map = {**self.color_map, **color_map}
            return Segmentation(class_names=class_indices, input_key=self.input_key, label_map_key=self.gt_key,
                                meta_keys=meta_keys, dontcare=self.dontcare, color_map=color_map)
        elif type(task) == Task:
            assert self.check_compat(task), f"cannot create compatible task between:\n\t{str(self)}\n\t{str(task)}"
            meta_keys = list(set(self.meta_keys + task.meta_keys))
            return Segmentation(class_names=self.class_indices, input_key=self.input_key, label_map_key=self.gt_key,
                                meta_keys=meta_keys, dontcare=self.dontcare, color_map=self.color_map)

    def __repr__(self):
        """Creates a print-friendly representation of a segmentation task."""
        color_map = {k: v.tolist() for k, v in self.color_map.items()}
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(class_names={repr(self.class_indices)}, input_key={repr(self.input_key)}, " + \
            f"label_map_key={repr(self.gt_key)}, meta_keys={repr(self.meta_keys)}, " + \
            f"dontcare={repr(self.dontcare)}, color_map={repr(color_map)})"
